#include "testing_flags.cuh"

#ifdef TEST_GROUP_SHARED

#include "testing_commons.cuh"

#include "tile/tile.cuh"
#include "vec/vec.cuh"

namespace group {
namespace shared {

void tests(test_data &results);

}
}

#endif