#include "testing_flags.cuh"

#ifdef TEST_GROUP_MEMORY_VEC_GLOBAL_TO_REGISTER

#include "testing_commons.cuh"

namespace group {
namespace memory {
namespace vec {
namespace global_to_register {

void tests(test_data &results);

}
}
}
}

#endif