/**
 * @file
 * @brief A collection of all of ThunderKittens prototypes, that can be filled in to easily build full kernels.
 */

#pragma once

#include "../include/kittens.cuh"

#include "common/common.cuh"
#include "lcf/lcf.cuh"
#include "lcsf/lcsf.cuh"