#include "testing_flags.cuh"

#ifdef TEST_WARP_SHARED_VEC_REDUCTIONS

#include "testing_commons.cuh"

namespace warp {
namespace shared {
namespace vec {
namespace reductions {

void tests(test_data &results);

}
}
}
}

#endif