/**
 * @file
 * @brief An aggregate header file for all the register types defined by ThunderKittens.
 */

#pragma once

#include "rv_layout.cuh"
#include "rt_base.cuh"
#include "rv.cuh"
#include "rt.cuh"

#include "crv.cuh"
#include "crt.cuh"

