#include "testing_flags.cuh"

#ifdef TEST_WARP_SHARED_TILE_MAPS

#include "testing_commons.cuh"

namespace warp {
namespace shared {
namespace tile {
namespace maps {

void tests(test_data &results);

}
}
}
}

#endif