#include "testing_flags.cuh"

#ifdef TEST_WARP_MEMORY_TILE_SHARED_TO_REGISTER

#include "testing_commons.cuh"

namespace warp {
namespace memory {
namespace tile {
namespace shared_to_register {

void tests(test_data &results);

}
}
}
}

#endif