#include "testing_flags.cuh"

#ifdef TEST_GROUP_SHARED_TILE_MAPS

#include "testing_commons.cuh"

namespace group {
namespace shared {
namespace tile {
namespace maps {

void tests(test_data &results);

}
}
}
}

#endif